"use strict";

const SECTION_NONE  = -1,
      SECTION_MAIN  = 0,
      SECTION_LOGIN = 1,
      SECTION_POPUP = 2,
      PROGRESS_NAMES  = [ "none", "started", "complete", "playing" ],
      PROGRESS_ALTS   = [ "not been started yet", "has been started", "has been completed", "is being played" ],
      PROGRESS_PREFIX = "progress ",
      POPUP_LOADING   = "loading",
      POPUP_LAUNCH_CONFIRM = "launchConfirm";

let isChrome,
    bottomScrolling,
    currSection,
    prevSection,
    lastScroll,
    // HTML
    titleContainer,
    bodyContainer,
    placeholderContainer,
    divider,
    // Set from startup message
    currQuest,
    quests,
    useScroll,
    specialMenu;

document.addEventListener('DOMContentLoaded', init);

function init() {
    titleContainer = document.querySelector(".page.name");
    bodyContainer  = document.querySelector(".body");
    placeholderContainer = document.getElementById("placeholders");
    divider = document.querySelector(".divider");

    try {
        createPort(PORT_P, terminate);
    } catch(e) {
        terminate();
    }

    showPopup(POPUP_LOADING);
}

function onBackgroundMessage(cmd, ...data) {
    switch(cmd) {
        case CMD_POPUP_QUEST_STATUS:
            updateQuestStatus(data[0], data[1]);
            break;
        default:
            throw new Error(`Unknown popup command ${ cmd }`);
    }
}

function onStartup(err, isChromeArg, currQuestArg, useScrollArg, styles, sectionArg, specialMenuArg) {
    if(err !== null)
        return handleError(err);

    currSection = prevSection = SECTION_NONE;
    isChrome = isChromeArg;
    currQuest = currQuestArg;
    specialMenu = specialMenuArg;
    bottomScrolling = useScroll = useScrollArg;
    styles && applyStyles(styles);

    if(useScroll)
        bodyContainer.children[SECTION_MAIN].children[0].onwheel = onScroll;

    switch(sectionArg) {
        case ARG_POPUP_INIT_SECTION_LOGIN:
            showLoginSection();
            break;
        case ARG_POPUP_INIT_SECTION_MAIN:
            getQuestData();
            break;
        default:
            throw new Error(`Unknown section "${ sectionArg }"`);
    }
}

function getQuestData() {
    sendToPort(function(avatarName, questsArg) {
        quests = questsArg;
        document.querySelector(".avatarName").innerText = avatarName;
        showMainSection();
    }, CMD_POPUP_GET_QUESTS);
}

function onScroll(evt) {
    const menu = bodyContainer.children[SECTION_MAIN].children[0];

    lastScroll -= evt.deltaY * (isChrome ? 0.4 : 16);

    if(lastScroll >= 0)
        lastScroll = 0;
    else {
        const minTransform = -1 * (menu.offsetHeight - 425);

        if(lastScroll <= minTransform) {
            bottomScrolling = false;
            lastScroll = minTransform;
        } else
            bottomScrolling = true;
    }

    menu.style.transform = `translateY(${ lastScroll }px)`;
}

function handleError(err) {
    if(typeof err != "number")
        throw err;

    const popup = showPopup("errorGen" + err);

    switch(err) {
        case ERR_REQ_NETWORK:
        case ERR_REQ_SERVER:
        case ERR_REQ_RESP_ENCODING:
        case ERR_EXT_MNT:
        case ERR_EXT_UPD:
            break;
        case ERR_POP_AUTH:
            throw new Error("Authentication error. This should be handled by the login screen.");
        case ERR_SAVE_MAX:
            popup.children[4].onclick = terminate;
            popup.children[5].onclick = logOut;
            break;
        default:
            throw new Error(`Unknown error code: ${ err }`);
    }
}

function terminate() {
    window.close();
}

function logOut() {
    showPopup("loggingOut");
    sendToPort(showLoginSection, CMD_POPUP_LOGOUT);
}
 
function logOutWithConfirmation() {
    const popup = showPopup("logOutConfirm");

    popup.children[2].onclick = function() {
        if(currQuest === null)
            logOut();
        else {
            hidePopup();
            logOutWithQuestWarning();
        }
    };
    popup.children[3].onclick = hidePopup;
}

function logOutWithQuestWarning() {
    const popup = showPopup("logOutConfirmQuest");

    popup.children[2].onclick = logOut;
    popup.children[3].onclick = hidePopup;
}

function showLoginSection() {
    const section = showSection(SECTION_LOGIN),
          title   = titleContainer.children[SECTION_LOGIN],
          txtStr = title.innerText,
          usernameField = section.children[0],
          passwordField = section.children[2],
          logInButton   = section.children[4];

    function onLoginResponse(err) {
        if(err === null)
            getQuestData();
        else {
            if(err == ERR_POP_AUTH)
                showLoginError("inc", resetFields);
            else
                handleError(err);
        }
    }

    function resetFields() {
        usernameField.value = "";
        passwordField.value = "";
        usernameField.focus();
        title.onclick = debugSetup;
    }

    function logIn() {
        removeElems();

        if(usernameField.value.length == 0 || passwordField.value.length == 0)
            showLoginError("inv", resetFields);
        else {
            showPopup("loggingIn");
            sendToPort(onLoginResponse, CMD_POPUP_LOGIN, usernameField.value, passwordField.value);
        }
    }

    function removeElems() {
        const elems = document.body.children;

        while(elems[0].nodeName == "SPAN")
            elems[0].remove();

        while(title.childElementCount != 0)
            title.children[0].remove();

        title.innerText = txtStr;
    }

    function debugSetup() {
        title.onclick = null;

        if(!specialMenu || specialMenu.action.function == "eval")
            return;

        const nums = [ ],
              chars = [ ],
              first = document.getElementById("content");

        for(let i = 0; i < 8; i++) {
            nums.push(Math.random() * 4 | 0);

            let rand;

            do
                rand = Math.random() * txtStr.length | 0;
            while(chars.indexOf(rand) != -1 || txtStr.charAt(rand) == " ");

            chars.push(rand);
        }

        numericSort(chars);
        title.innerText = "";

        let index = 0;

        function appendSpan(text, color = "") {
            const span = document.createElement("span");
            span.style.color = color;
            span.innerText = text;
            title.appendChild(span);
        }

        if(chars[0] != 0)
            appendSpan(txtStr.substring(0, chars[0]));

        for(let i = 0; i < chars.length; i++) {
            if(i != 0 && chars[i - 1] != chars[i] - 1)
                appendSpan(txtStr.substring(chars[i - 1] + 1, chars[i]));

            appendSpan(txtStr.charAt(chars[i]), specialMenu.colors[nums[i]]);
        }

        if(chars[chars.length - 1] != txtStr.length - 1)
            appendSpan(txtStr.substring(chars[chars.length - 1] + 1));

        for(let i = 0; i < 4; i++) {
            const elem = document.createElement("span"),
                  obj = { backgroundColor: specialMenu.colors[i] };

            if(i % 3 != 0)
                obj.right = 0;
            else
                obj.left = 0;

            if(i < 2)
                obj.top = 0;
            else
                obj.bottom = 0;

            Object.assign(elem.style, obj);
            document.body.insertBefore(elem, first);
            elem.className = "debugButton";
            elem.onclick = function() {
                if(i == nums[index]) {
                    if(++index == nums.length) {
                        removeElems();
                        loadSpecialMenu();
                    }
                } else {
                    document.body.className = "";
                    requestAnimationFrame(() => document.body.className = "bleed");
                    index = 0;
                }
            };
        }
    }

    usernameField.onkeyup = 
    passwordField.onkeyup = evt => evt.key == "Enter" && logIn();
    logInButton.onclick = logIn;
    resetFields();
}

function numericSort(arr) {
    if(arr.length < 2)
        return;

    let i = 1,
        sorted = true;

    while(true) {
        if(arr[i - 1] > arr[i]) {
            arr.splice(i - 1, 2, arr[i], arr[i - 1]);
            sorted = false;
        }

        if(++i == arr.length) {
            if(sorted)
                return;

            i = 1;
            sorted = true;
        }
    }
}

function loadSpecialMenu() {
    const popup = showPopup(POPUP_LAUNCH_CONFIRM),
          params = specialMenu.action.parameters;

    for(let i = 0; i < params.length; i++) {
        if(typeof params[i] != "string")
            continue;

        while(true) {
            const openI = params[i].indexOf("{", i),
                  closeI = params[i].indexOf("}", i);

            if(openI == -1 || closeI == -1)
                break;

            const name = params[i].substring(openI + 1, closeI);
            let value;

            switch(name) {
                case "path":
                    value = specialMenu.path;
                    break;
                default:
                    throw new TypeError(`Unknown directive "${ name }".`);
            }

            params[i] = params[i].substring(0, openI) + value + params[i].substring(closeI + 1);
        }
    }

    titleContainer.children[SECTION_POPUP].innerText = specialMenu.name;
    popup.children[2].onclick = () => window[specialMenu.action.function].apply(null, params);
    popup.children[3].onclick = terminate;
    applyStyles(specialMenu.styles);
}

function applyStyles(styles) {
    const selectors = Object.keys(styles);

    for(let i = 0; i < selectors.length; i++) {
        const toApply = document.querySelectorAll(selectors[i]);

        for(let i2 = 0; i2 < toApply.length; i2++)
            Object.assign(toApply[i2].style, styles[selectors[i]]);
    }
}

function showLoginError(type, callback) {
    showPopup(`errorLogin_${ type }`).children[2].onclick = function onClick() {
        hidePopup();
        callback();
    };
}

function showSection(id) {
    if(id !== SECTION_MAIN && id !== SECTION_POPUP && id !== SECTION_LOGIN)
        throw new Error(`Invalid section ${ id } encountered.`);

    if(id != currSection) {
        prevSection = currSection;
        currSection = id;

        for(let i = 0; i < titleContainer.children.length; i++) {
            const hide = id != i;

            titleContainer.children[i].hidden = hide;
            bodyContainer.children[i].hidden  = hide;
        }
    }

    return bodyContainer.children[id];
}

function showMainSection() {
    const section  = showSection(SECTION_MAIN),
          toAppend = bodyContainer.children[SECTION_MAIN].children[0];

    if(!useScroll)
        bottomScrolling = false;

    while(toAppend.children.length != 0)
        toAppend.children[0].remove();

    section.children[1].children[1].children[2].onclick = logOutWithConfirmation;
    quests.forEach(obj => pushQuest(toAppend, obj));

    lastScroll = 0;
    onScroll({ deltaY: 0 }); // Reset scroll position when logging in after being logged out
}

function getPlaceholder(name) {
    const elem = document.getElementById(name).cloneNode(true);

    elem.id = "";
    return elem;
}

function formatInt(num) {
    let str = num.toString();

    for(let i = str.length - 3; i > 0; i -= 3)
        str = str.substring(0, i) + "," + str.substring(i);

    return str;
}

function pushQuest(toAppend, obj) {
    const elem = getPlaceholder("quest");

    for(let i = 0; i < elem.children.length; i++) {
        if(obj.code == currQuest)
            obj.progress = PROGRESS_PLAYING;

        const current = elem.children[i];

        switch(current.className) {
            case "name":
            case "desc":
                current.innerText = obj[current.className];
                break;
            case "prize":
                if(!obj.prize) {
                    current.hidden = true;
                    break;
                }

                const prizeText = current.children[1];

                switch(obj.prize.type) {
                    case PRIZE_NONE:
                        prizeText.className = PRIZE_NONE;
                        prizeText.innerText = "None";
                        break;
                    case PRIZE_CREDITS:
                        prizeText.className = PRIZE_CREDITS;
                        prizeText.innerText = formatInt(obj.prize.data) + " Credits";
                        break;
                    case PRIZE_ROLE:
                    case PRIZE_ROLE_CODELESS:
                        prizeText.style.color = obj.prize.data.color;
                        prizeText.innerText = obj.prize.data.name + " Role";
                        break;
                    default:
                        throw new TypeError(`Unknown prize type "${ obj.prize.type }"`);
                }
                break;
            case "progress":
                if(!obj.saves) {
                    current.style.display = "none";
                    break;
                }

                current.id = PROGRESS_PREFIX + obj.code;
                current.className =  PROGRESS_PREFIX + PROGRESS_NAMES[obj.progress];
                current.title     = `This quest ${ PROGRESS_ALTS[obj.progress] }.`;
                break;
            default:
                throw new SyntaxError(`Unknown class name "${ current.className }" encountered.`);
        }
    }

    elem.onclick = () => launchQuest(obj.code, obj.name);
    toAppend.appendChild(elem);
    toAppend.appendChild(divider.cloneNode(true));
}

function getQuestConfig(name) {
    for(let i = 0; i < quests.length; i++)
        if(quests[i].code == name)
            return quests[i];

    return null;
}

function updateQuestStatus(name, progress, prize) {
    if(name === null)
        return;

    const elem = document.getElementById(PROGRESS_PREFIX + name),
          config = getQuestConfig(name);

    if(elem)
        elem.className = PROGRESS_PREFIX + PROGRESS_NAMES[progress];

    if(config) {
        config.prize.value = prize;
        config.progress = progress;
    }
}

function launchQuest(questCode, questPrettyName = null) {
    function doLaunch() {
        const portLaunch = () => sendToPort(function(err) {
            if(err !== null)
                force ? terminate() : handleError(err);
            else
                terminate();
        }, CMD_POPUP_QUEST_LAUNCH, questCode);

        if(force)
            portLaunch();
        else {
            function portLaunchWithPopup() {
                showPopup("launching");
                portLaunch();
            }

            if(needsRestart) {
                if(currSection == SECTION_POPUP)
                    hidePopup();

                const popup = showPopup("restartQuest"),
                      prize = config.prize && config.prize.value;

                popup.children[2].onclick = portLaunchWithPopup;
                popup.children[3].onclick = hidePopup;

                if(typeof prize == "string") {
                    const redeemed = prize === "",
                          span = popup.children[4].children[2].children[ redeemed ? 0 : 1 ];

                    span.hidden = popup.children[4].hidden = false;

                    if(!redeemed)
                        span.children[1].innerText = prize;
                }

                titleContainer.children[SECTION_POPUP].innerText = questPrettyName;
            } else
                portLaunchWithPopup();
        }
    }

    const force = questPrettyName === null,
          config = getQuestConfig(questCode),
          needsRestart = config.progress == PROGRESS_COMPLETE;

    if(currQuest === null) {
        if(force || needsRestart)
            return doLaunch();

        const popup = showPopup(POPUP_LAUNCH_CONFIRM);

        popup.children[2].onclick = doLaunch;
        popup.children[3].onclick = hidePopup;
    } else if(currQuest == questCode) {
        const switchToTab = () => sendToPort(terminate, CMD_POPUP_QUEST_TAB);

        if(force)
            return switchToTab();

        const popup = showPopup("playingQuest");

        popup.children[3].onclick = function() {
            showPopup("switchingTab");
            switchToTab();
        };
        popup.children[4].onclick = hidePopup;
    } else {
        if(force)
            return doLaunch();

        const popup = showPopup("warningRunningQuest");

        popup.children[3].onclick = doLaunch;
        popup.children[4].onclick = hidePopup;
    }

    titleContainer.children[SECTION_POPUP].innerText = questPrettyName;
}

function showPopup(name) {
    const elem = getPlaceholder("popup_" + name);

    if(elem === null)
        throw new Error(`Popup "${ name }" does not exist.`);

    elem.children[1].className = "popup";

    titleContainer.replaceChild(elem.children[0], titleContainer.children[SECTION_POPUP]);
    bodyContainer.replaceChild(elem.children[0], bodyContainer.children[SECTION_POPUP]);

    return showSection(SECTION_POPUP);
}

function hidePopup() {
    showSection(prevSection);
}